#ifndef MGML_MATH_EXAMPLE_GUARDIAN
#define MGML_MATH_EXAMPLE_GUARDIAN


//#ifndef MGML_GUARDIAN
//	#include "../CSL/MGML_GLOBALS.h"
//#endif

#ifndef MGML_GUARDIAN
	#include "../CSL/MGML.h"
#endif

using namespace MGML;

///////////////////////////////////// + ////////////////////////////////////////////
__global__ void test_kernel(float4 *in_a, float4* in_b, float4* out_res)
{
  unsigned int tid = blockDim.x * blockIdx.x + threadIdx.x;

  float4 a = in_a[tid];
  float4 b = in_b[tid];
 
  float4 res = make_float4(a.x+b.x, a.y+b.y, a.z+b.z, a.w+b.w);

  out_res[tid] = res;
}


///////////////////////////////////// dot ////////////////////////////////////////////
__global__ void test_kernel_dot(float4 *in_a, float4* in_b, float4* out_res)
{
  unsigned int tid = blockDim.x * blockIdx.x + threadIdx.x;

  float4 a = in_a[tid];
  float4 b = in_b[tid];
 
  float dot = a.x*b.x + a.y*b.y + a.z*b.z;
  float4 res = make_float4(dot, dot, dot, dot);

  out_res[tid] = res;
}

///////////////////////////////////// ray triangle intersection ////////////////////////////////////////////


__device__ float dot(float4 a, float4 b) 
{ 
  return a.x*b.x + a.y*b.y + a.z*b.z;
}

__device__ float4 cross3(float4 a, float4 b)
{
  return make_float4(a.y*b.z - a.y*b.z,
			              a.z*b.x - a.z*b.x,
			              a.x*b.y - a.x*b.y, 1);
}

__device__ float4 operator-(float4 a, float4 b)
{
  return make_float4(a.x-b.x ,a.y-b.y, a.z-b.z, 1);
}

struct MyRay
{
    float4 pos;
    float4 dir;
};

__device__ bool RayTriangleIntersectionTest(const MyRay& ray, float4 A, float4 B, float4 C, float4* bar)
{
  

  float4 edge1, edge2,pvec,tvec,qvec;
  float  det,inv_det;

  edge1 = B - A;
  edge2 = C - A;

  pvec  = cross3(ray.dir,edge2);
 
  det  =  dot(edge1,pvec);
  //culling is disabled 
  //  enable cull next [if] must be 
   // if(det < EPSILON_E6M)
  if (det > -EPSILON_E6M && det < EPSILON_E6M)
     return false;
  inv_det = 1.0f / det;
 
  tvec = ray.pos - A;

  bar->y = dot(tvec,pvec)*inv_det;
  if (bar->y < -0.001f || bar->y > 1.001f)
   return false;
 
  qvec = cross3(tvec, edge1);
  bar->z = dot(ray.dir,qvec)*inv_det;
  if (bar->z < -0.001f || bar->y + bar->z > 1.001f)
     return false;

  bar->x = 1.0f - bar->z - bar->y;
  return true;
}


__global__ void test_intersection_kernel(float4 *in_a, float4* in_b, float4* out_res)
{
  unsigned int tid = blockDim.x * blockIdx.x + threadIdx.x;

  float4 A = in_a[tid];
  float4 B = in_b[tid];
  float4 C = B;
  float4 bar;
  float4 res;
  MyRay ray;
  ray.pos = A;
  ray.dir = B;
  
  if(RayTriangleIntersectionTest(ray,A,B,C,&bar))
  {
    res.x = 1;
    res.y = 1;
    res.z = 1;
    res.w = 1;
  }
  else
  {
    res.x = 0;
    res.y = 0;
    res.z = 0;
    res.w = 0;
  }

  out_res[tid] = res;
}



///////////////////////////////////// + ////////////////////////////////////////////
__global__ void test_kernel2(vec4f *in_a, vec4f* in_b, vec4f* out_res)
{
  unsigned int tid = blockDim.x * blockIdx.x + threadIdx.x;

  vec4f a = in_a[tid];
  vec4f b = in_b[tid];
 
  vec4f res = a + b;

  out_res[tid] = res;
}

///////////////////////////////////// dot ////////////////////////////////////////////
__global__ void test_kernel_dot2(vec4f *in_a, vec4f* in_b, vec4f* out_res)
{
  unsigned int tid = blockDim.x * blockIdx.x + threadIdx.x;

  vec4f a = in_a[tid];
  vec4f b = in_b[tid];
 
  float dot = a*b;
  vec4f res;
  res.set(dot,dot,dot,dot);

  out_res[tid] = res;
}

///////////////////////////////////// + ////////////////////////////////////////////
__global__ void test_intersection_kernel2(vec4f *in_a, vec4f* in_b, vec4f* out_res)
{
  unsigned int tid = blockDim.x * blockIdx.x + threadIdx.x;

  vec4f A = in_a[tid];
  vec4f B = in_b[tid];
  vec4f C = B;
  vec4f bar;
  vec4f res;
  Ray4f ray;
  ray.pos = A;
  ray.dir = B;

  if(MGML_MATH::Intersect<Ray4f,Triangle4f>::exec(ray,A,B,C,&bar))
    res.set(1,1,1,1);
  else
    res.set(0,0,0,0);

  out_res[tid] = res;
}

#endif 
